-------------------------------------------------------------------
-- Tiny PCI target for KSI Labs, LLC CENTIPEDE-PCI board. Very   --
-- basic, just essential functionality -- no bursting, no parity --
-- checking, no cache lines, no #SERR, no Target Abort/Retry,    --
-- only one 4K memory space BAR, only 32-bit access etc.         --
-------------------------------------------------------------------

--------------------------------------------------------------------------
--			LICENSE and Terms of Use.			--
--									--
-- First of all, this is NOT GPL or Free Software. It is open so you	--
-- have the VHDL source code. There is also the source code for the	--
-- Windows Engine for Mach3 CNC software available from KSI Labs, LLC	--
-- website, http://www.ksilabs.com.					--
--									--
-- You are free to use this particular VHDL code, make modifications to	--
-- customize it for your own needs, fix inevitable bugs, port it to 	--
-- other CNC software and do anything else as long as your derivative	--
-- work that is based on this code or uses parts or pieces thereof is	--
-- made to be used with any KSI Labs, LLC products.			--
--									--
-- We do NOT grant you a permission to use any parts of this code in	--
-- any products that are designed for products not manufactured or	--
-- distributed by KSI Labs, LLC. We might give such a permission on	--
-- case-by-case basis but it is not legal to use this code or bits and	--
-- pieces thereof with products not coming from KSI Labs, LLC without	--
-- our explicit written permission.					--
--									--
-- The original copyright notices must be included in all derivative	--
-- works. You do NOT have to ask our permission to make any changes to	--
-- this code and you are not under obligation to give us any of your	--
-- work back as long as you are using it for your own needs.		--
--									--
-- However we do NOT give our permission to distribute your derivative	--
-- work unless all original copyrights are retained, all the conditions	--
-- of this license are met (use only with KSI Labs, LLC products) and	--
-- full source code for such derivative work is available for free to	--
-- everybody without any further conditions including but not limited	--
-- to signing any kinds of NDA. We explicitely deny any permission to	--
-- distribute any derivative work based on this code with closed source	--
-- binary-only fragments or include bits or pieces of this code in such	--
-- products. That essentially means that it is NOT legal to use it with --
-- ANY proprietary HDL megafunctions, megacores, etc. if the resulting	--
-- product is distributed. It IS legal to use whatever you want if the	--
-- resulting product is made for use with a KSI Labs, LLC product AND	--
-- that resulting product is not distributed. Exclusions can be granted --
-- by KSI Labs, LLC on case-by-case basis.				--
--------------------------------------------------------------------------

library ieee;
use ieee.std_logic_1164.all;

entity pci32ttiny is
generic (
	vendorID	: std_logic_vector(15 downto 0)	:= x"feff";
	deviceID	: std_logic_vector(15 downto 0)	:= x"0001";
	revisionID	: std_logic_vector(7 downto 0)	:= x"00";
	subsystemID	: std_logic_vector(15 downto 0)	:= x"0202";
	subsystemvID	: std_logic_vector(15 downto 0)	:= x"1172";
	classcodeID	: std_logic_vector(23 downto 0)	:= x"078000"
);
port (
	-- General 
	pci_clk		: in std_logic;
	pci_rst		: in std_logic;
    
	-- PCI target 32bits
	pci_ad		: inout std_logic_vector(31 downto 0);
	pci_cbe		: in std_logic_vector(3 downto 0);
	pci_par		: out std_logic;  
	pci_frame	: in std_logic;
	pci_irdy	: in std_logic;
	pci_trdy	: out std_logic;
	pci_devsel	: out std_logic;
	pci_stop	: out std_logic;
	pci_idsel	: in std_logic;
	pci_perr	: out std_logic;
	pci_serr	: out std_logic;
	pci_inta	: out std_logic;
      
	-- Local bus
	lb_adr		: out std_logic_vector(11 downto 0);     
	lb_data_in	: out std_logic_vector(31 downto 0);
	lb_data_out	: in std_logic_vector(31 downto 0);
	lb_nbe		: out std_logic_vector(3 downto 0);
	lb_nint		: in std_logic;
	lb_rd		: out std_logic;
	lb_wr		: out std_logic
);
end pci32ttiny;


architecture guts of pci32ttiny is
	constant CLASSCODE_REG	: std_logic_vector(23 downto 0) := classcodeID;   
	constant REVISIONID_REG	: std_logic_vector(7 downto 0) := revisionID;	-- PR00=80,PR1=81...
	constant HEADERTYPE_REG	: std_logic_vector(7 downto 0) := x"00";		
	constant DEVSELTIM_BITS	: std_logic_vector(1 downto 0) := b"01";		-- DEVSEL TIMing (bits) medium speed
	constant VENDORID_REG	: std_logic_vector(15 downto 0) := vendorID;	
	constant DEVICEID_REG	: std_logic_vector(15 downto 0) := deviceID;	
	constant SUBSYSID_REG	: std_logic_vector(15 downto 0) := subsystemID;	
	constant SUBSYSVID_REG	: std_logic_vector(15 downto 0) := subsystemvID;	
	constant INTPIN_REG	: std_logic_vector(7 downto 0):= x"01";		-- INTA#

	signal bar0		: std_logic_vector(31 downto 12);
	signal mem_en		: std_logic;
	signal pci_ale		: std_logic;
	signal cfg_adr		: std_logic_vector(5 downto 0);
	signal cfg_sel		: std_logic;
	signal lb_sel		: std_logic;
	signal pci_data_oe	: std_logic;
	signal pci_par_oe	: std_logic;	
	signal cfg_wr		: std_logic;
	signal cfg_rd		: std_logic;
	signal data_to_pci	: std_logic_vector(31 downto 0);
	signal data_from_pci	: std_logic_vector(31 downto 0);

	-- PCI FSM state signals
	type pci_fsm_state is (IDLE, BUSY, DATA1, DATA2, TURNARND, IGNORE);
	signal fsm_cur 		: pci_fsm_state;
	signal fsm_nxt 		: pci_fsm_state;
	-- Other PCI FSM related signals
	signal s_idle		: std_logic;
	signal s_data1		: std_logic;
	signal s_data2		: std_logic;
	signal s_data1_nxt	: std_logic;
	signal s_turnarnd_nxt	: std_logic;
	signal s_turnarnd	: std_logic;
	signal ndevsel_nxt	: std_logic;
	signal ntrdy_nxt	: std_logic;
	signal nstop_nxt	: std_logic;
	signal devsel		: std_logic;
	signal trdy		: std_logic;
	signal stop		: std_logic;
	signal ack		: std_logic;
	signal tgt_oe		: std_logic;

	-- Latches for PCI address, command, and IDSEL
  	signal adr_latch	: std_logic_vector(31 downto 0);
  	signal cmd_latch	: std_logic_vector(3 downto 0);
  	signal idsel_latch	: std_logic;

  	-- CFG registers data output, goes to data_to_pci->PCI
	signal cfg_data_out	: std_logic_vector(31 downto 0);
	-- CFG registers addresses detected/write enables
	signal stcmd_reg_sel	: std_logic;
	signal bar0_reg_sel	: std_logic;
	signal int_reg_sel	: std_logic;

	-- INTERRUPT LINE register 
  	signal INTLINE_REG	: std_logic_vector(7 downto 0);
	-- COMMAND register bits
	signal MEMSPACEEN_BIT	: std_logic;
	signal PERREN_BIT	: std_logic;	-- NO effect
	signal SERREN_BIT	: std_logic;	-- NO effect
	-- STATUS register bits
	signal TABORTSI_BIT	: std_logic;	-- NOT updated
	signal SERRSI_BIT	: std_logic;	-- NOT updated
	signal PERRDT_BIT	: std_logic;	-- NOT updated
	-- BAR0 register bits
	signal BAR0_REG		: std_logic_vector(31 downto 0);

	-- Parity generator signals
  	signal dout		: std_logic_vector(31 downto 0);
  	signal pardat		: std_logic;
  	signal parcbe		: std_logic;
  	signal par_int		: std_logic;
  	signal par_out		: std_logic;

begin

	-----------------------
	-- PCI FSM, top half --
	-----------------------
	pci_fsm_top : process (pci_rst, pci_clk, fsm_nxt)
	begin	
		if (pci_rst='0') then
			fsm_cur <= IDLE;
		elsif (rising_edge(pci_clk)) then
			fsm_cur <= fsm_nxt; 
		end if;
	end process;


	---------------------------------------------------------------------
	-- PCI FSM, bottom half. Note: We DO support Fast Back-to-Back     --
	-- (going TURNARND->BUSY when master keeps FRAME asserted) however --
	-- it is not used and we do not show it in our capabilities.       --
	-- We do NOT support retries, our registers are always ready.      --
	---------------------------------------------------------------------
  	pci_fsm_bottom : process (fsm_cur, pci_frame, pci_irdy, cfg_sel, lb_sel, ack)
	begin
		ndevsel_nxt <= '1';
		ntrdy_nxt <= '1';	
		nstop_nxt <= '1';	
		case fsm_cur is
    			when IDLE =>
    				if (pci_frame='0') then	
    					fsm_nxt <= BUSY; 	
				else
					fsm_nxt <= IDLE;
				end if;		
			when BUSY =>
				if ((cfg_sel or lb_sel)='0') then
					fsm_nxt <= IGNORE;
				else
					fsm_nxt <= DATA1;
					ndevsel_nxt <= '0'; 
				end if;
			when DATA1 =>
				if (ack='1') then	
					if (pci_frame='0') then
						nstop_nxt <= '0';	
					end if;
					fsm_nxt <= DATA2;
					ndevsel_nxt <= '0'; 					
					ntrdy_nxt <= '0';	
				else
					fsm_nxt <= DATA1;
					ndevsel_nxt <= '0'; 					
				end if;		
			when DATA2 => 
				fsm_nxt <= TURNARND;
			when TURNARND =>
				if (pci_frame='0') then
					fsm_nxt <= BUSY;
				else
					fsm_nxt <= IDLE;
				end if;				
			when IGNORE =>
				if (pci_frame='0') then
					fsm_nxt <= IGNORE;
				else
					fsm_nxt <= IDLE;
				end if;				
		end case;
	end process;	


	ack <= '1' when ((lb_sel='1' or cfg_sel='1') and pci_irdy='0') else '0';


	-- FSM state signals
	s_idle <= '1' when (fsm_cur=IDLE) else '0';
	s_data1 <= '1' when (fsm_cur=DATA1) else '0';
	s_data2 <= '1' when (fsm_cur=DATA2) else '0';
	s_turnarnd <= '1' when (fsm_cur=TURNARND) else '0';
	s_data1_nxt <= '1' when (fsm_nxt=DATA1) else '0';	
	s_turnarnd_nxt <= '1' when (fsm_nxt=TURNARND) else '0';


	-- PCI data OE
	pcidat_oe : process (pci_rst, pci_clk, cmd_latch(0), s_data1_nxt, s_turnarnd_nxt)
	begin
		if (pci_rst='0') then 
			pci_data_oe <= '0';
		elsif (rising_edge(pci_clk)) then 
			if (s_data1_nxt='1' and cmd_latch(0)='0') then
				pci_data_oe <= '1';
			elsif (s_turnarnd_nxt='1') then
				pci_data_oe <= '0';
			end if;			
		end if;
	end process;


	-- PCI parity OE (parity is 1 clock after data valid)
	pci_parity_oe : process (pci_clk, pci_rst, pci_data_oe)
	begin
		if (pci_rst='0') then
			pci_par_oe <= '0';
		elsif (rising_edge(pci_clk)) then
			pci_par_oe <= pci_data_oe;
		end if;
	end process;


	-- Target s/t/s signals OE
	tgt_oe_proc : process (pci_rst, pci_clk, s_data1_nxt, s_turnarnd)
	begin
		if (pci_rst='0') then 
			tgt_oe <= '0';
		elsif (rising_edge(pci_clk)) then 
			if (s_data1_nxt='1') then
				tgt_oe <= '1';
			elsif (s_turnarnd='1') then
				tgt_oe <= '0';
			end if;			
		end if;
	end process;
		

	-- PCI outputs	
	pci_outp_proc : process (pci_rst, pci_clk, ndevsel_nxt, ntrdy_nxt, nstop_nxt)
	begin
		if (pci_rst='0') then 
			devsel <= '1';
			trdy <= '1';
			stop <= '1';
		elsif (rising_edge(pci_clk)) then 
			devsel <= ndevsel_nxt;
			trdy <= ntrdy_nxt;
			stop <= nstop_nxt;	
		end if;		
	end process;


	pci_devsel <= devsel when (tgt_oe='1') else 'Z';
	pci_trdy <= trdy when (tgt_oe='1') else 'Z';
	pci_stop <= stop when (tgt_oe='1') else 'Z';

	
	-- Load enable for PCI address. s_turnarnd for Fast Back2Back
	pci_ale <= '1' when (pci_frame='0' and (s_idle='1' or s_turnarnd='1')) else '0';


	-- Latch PCI address/cmd and IDSEL# at address phase
	pci_ad_cmd_latch : process (pci_rst, pci_clk, pci_ad, pci_cbe, pci_idsel, pci_ale)
	begin
		if (pci_rst='0') then 
			adr_latch <= (others => '1');
			cmd_latch <= (others => '1');
			idsel_latch <= '0';
		elsif (rising_edge(pci_clk)) then 
			if (pci_ale='1') then
				adr_latch <= pci_ad;
				cmd_latch <= pci_cbe;
				idsel_latch <= pci_idsel;
			end if;
		end if;
	end process;


	-- Local bus access
	lb_sel <= '1' when ((mem_en='1') and
			(adr_latch(31 downto 12)=bar0(31 downto 12)) and
			(adr_latch(1 downto 0)="00") and
			(cmd_latch(3 downto 1)="011"))
		else '0';


	-- CFG space access
	cfg_sel <= '1' when ((idsel_latch='1') and
			(adr_latch(1 downto 0)="00" ) and 
			(cmd_latch(3 downto 1)="101"))
		else '0';


	-- CFG registers rd/wr signals
	cfg_wr <= '1' when (cfg_sel='1' and cmd_latch(0)='1' and s_data2='1') else '0';
	cfg_rd <= '1' when (cfg_sel='1' and cmd_latch(0)='0' and (s_data1='1' or s_data2='1')) else '0';


	-- Local bus registers rd/wr signals
	lb_wr <= '1' when (lb_sel='1' and cmd_latch(0)='1' and s_data2='1') else '0';
	lb_rd <= '1' when (lb_sel='1' and cmd_latch(0)='0' and (s_data1='1' or s_data2='1')) else '0';


	-- Local bus address
	lb_adr <= adr_latch(11 downto 2) & "00";


	-- CFG space address
	cfg_adr <= adr_latch(7 downto 2);


	-- 12 LSB are fixed '0' in BAR0 (4K address range)
	BAR0_REG(11 downto 0) <= (others => '0');


	-- CFG registers "chip selects"
	stcmd_reg_sel <= '1' when (cfg_adr(5 downto 0)=b"000001" ) else '0';
	bar0_reg_sel <= '1' when (cfg_adr(5 downto 0)=b"000100" ) else '0';
	int_reg_sel <= '1' when (cfg_adr(5 downto 0)=b"001111" ) else '0';


	-- COMMAND register write
	cmd_reg_wr : process (pci_clk, pci_rst, cfg_wr, stcmd_reg_sel, pci_cbe, data_from_pci)
	begin
		if (pci_rst='0') then
			MEMSPACEEN_BIT <= '0';
			PERREN_BIT <= '0';
			SERREN_BIT <= '0';			
		elsif (rising_edge(pci_clk)) then
			if (stcmd_reg_sel='1' and cfg_wr='1') then
				if (pci_cbe(0)='0') then
					MEMSPACEEN_BIT <= data_from_pci(1);
					PERREN_BIT <= data_from_pci(6);
				end if;
				if (pci_cbe(1)='0') then
					SERREN_BIT <= data_from_pci(8);
				end if;
			end if;
		end if;
	end process;


	-- We do NOT support those, fixed at '0'
	TABORTSI_BIT <= '0';
	SERRSI_BIT <= '0';
	PERRDT_BIT <= '0';


        -- INTERRUPT register write
        int_reg_wr : process (pci_clk, pci_rst, int_reg_sel, pci_cbe, data_from_pci)
        begin
        	if (pci_rst='0') then
			INTLINE_REG <= (others => '0');
		elsif (rising_edge(pci_clk)) then
			if (int_reg_sel='1' and cfg_wr='1') then
				if (pci_cbe(0)='0') then
					INTLINE_REG <= data_from_pci(7 downto 0);
				end if;
			end if;
		end if;
	end process;


	-- BAR0 register write (3 upper bytes, [31:12] for 4K area)
	bar0_reg_wr : process (pci_clk, pci_rst, bar0_reg_sel, pci_cbe, data_from_pci)
	begin
		if (pci_rst='0') then
			BAR0_REG(31 downto 12) <= (others => '0');
		elsif (rising_edge(pci_clk)) then
			if (bar0_reg_sel='1' and cfg_wr='1') then
				if (pci_cbe(3)='0') then
					BAR0_REG(31 downto 24) <= data_from_pci(31 downto 24);
				end if;
				if (pci_cbe(2)='0') then
					BAR0_REG(23 downto 16) <= data_from_pci(23 downto 16);
				end if;
				if (pci_cbe(1)='0') then
					BAR0_REG(15 downto 12) <= data_from_pci(15 downto 12);
				end if;
			end if;
		end if;
	end process;


	-- Read CFG registers
	cfg_reg_rd : process (cfg_adr, PERRDT_BIT, SERRSI_BIT, TABORTSI_BIT,
				SERREN_BIT, PERREN_BIT, MEMSPACEEN_BIT,
				BAR0_REG, INTLINE_REG, cfg_rd)
	begin
		if (cfg_rd='1') then
			case cfg_adr is
				when b"000000" => 
					cfg_data_out <= DEVICEID_REG & VENDORID_REG;
				when b"000001" => 
					cfg_data_out <= PERRDT_BIT & SERRSI_BIT &
						b"00" & TABORTSI_BIT & DEVSELTIM_BITS &
						b"000000000" & b"0000000" & SERREN_BIT &
						b"0" & PERREN_BIT & b"0000" & MEMSPACEEN_BIT & b"0";
				when b"000010" => 
					cfg_data_out <= CLASSCODE_REG & REVISIONID_REG;
				when b"000100" => 
					cfg_data_out <= BAR0_REG;
				when b"001011" => 
					cfg_data_out <= SUBSYSID_REG & SUBSYSVID_REG;
				when b"001111" => 
					cfg_data_out <= b"0000000000000000" & INTPIN_REG & INTLINE_REG;
				when others => 
					cfg_data_out <= (others => '0');
			end case;
		else
			cfg_data_out <= (others => '0');
		end if;
	end process;


	bar0 <= BAR0_REG(31 downto 12);
	mem_en <= MEMSPACEEN_BIT;


	-- We only GENERATE parity on output FROM target to PCI.
	-- We do NOT check parity.
	dout <= data_to_pci;


	-- Data parity
	pardat 	<= dout(0) xor dout(1) xor dout(2) xor dout(3) xor dout(4) xor
			dout(5) xor dout(6) xor dout(7) xor dout(8) xor
			dout(9) xor dout(10) xor dout(11) xor dout(12) xor
			dout(13) xor dout(14) xor dout(15) xor dout(16) xor
			dout(17) xor dout(18) xor dout(19) xor dout(20) xor
			dout(21) xor dout(22) xor dout(23) xor dout(24) xor
			dout(25) xor dout(26) xor dout(27) xor dout(28) xor
			dout(29) xor dout(30) xor dout(31);

						
	-- CBE parity
	parcbe 	<= pci_cbe(0) xor pci_cbe(1) xor pci_cbe(2) xor pci_cbe(3); 

	
	-- Full parity
	par_int <= pardat xor parcbe;


	-- Delay parity one clock
	par_dly : process (pci_clk, par_int, par_out)
	begin
		if (rising_edge(pci_clk)) then
			par_out <= par_int;
		end if;
	end process;


	-- PCI par pin
	pci_par <= par_out when (pci_par_oe='1') else 'Z';


	-- PCI AD[0:31]
	data_to_pci <= cfg_data_out when (cfg_rd='1') else lb_data_out;
	pci_ad <= data_to_pci when (pci_data_oe='1') else (others => 'Z');
	data_from_pci <= pci_ad;
	lb_data_in <= data_from_pci;


	lb_nbe <= pci_cbe;


	-- Not used, open drain
	pci_perr <= 'Z';
	pci_serr <= 'Z';

	
	-- PCI inta pin, open drain
	pci_inta <= '0' when (lb_nint='0') else 'Z';
end guts;
